# $Revision: 1.6 $
# $Date: 2003/03/25 08:08:24 $
package Net::IMAP::Server::RSSBridge;

=head1 NAME

Net::IMAP::Server::RSSBridge - Provides a backend to Net::IMAP::Server for reading RSS feeds

=cut

use strict;
use warnings;
use Data::Dumper;

use POSIX;

use Carp;
use Mail::Message;

use XML::RSS;
use Time::Local;
use LWP::UserAgent;

use HTML::RSSAutodiscovery;
use HTML::TreeBuilder;
use HTML::FormatText;

our $VERSION			= '0.01';
our $REVISION			= (qw$Revision: 1.6 $)[1];


my @mailboxes			= qw(INBOX);
my $DEFAULT_URL			= q(http://kasei.us/feeds/weblog.xml);
my $CACHE_MINUTES		= 120;
our $debug				= 0;

=head1 METHODS

=over 4

=item C<< new ( $username, $password ) >>

Acts as the backend's constructor. Returns a backend object representing a specific user's data.

=cut
sub new {
	my $proto	= shift;
	my $class	= ref($proto) || $proto;
	my $server	= shift;
	
	my $ua		= new LWP::UserAgent ( agent => 'mangalavid-RSSIMAPBridge/0.01' );
	my $self	= bless(
		{
			server		=> $server,
			ua			=> $ua,
			seperator	=> '#',
			_feed_cache	=> {}
		}, $class );
	
	return $self;
}

################################################################################

=item C<< fetch_seperator () >>

Returns the one character mailbox hierarchy seperator.

=cut
sub fetch_seperator {
	my $self	= shift;
	return $self->{ 'SEPERATOR' };
}

=item C<< list_mailboxes ( <$pattern> ) >>

Returns a LIST whose elements are the names of mailboxes. If $pattern is specified, it is used as a regular expression and only those mailboxes whose name matches the pattern are returned.

=cut
sub list_mailboxes {
	my $self	= shift;
	my $list	= $self->server->user->{ 'subscriptions' };
	if (ref($list)) {
		if (scalar(@_)) {
			my $pattern	= shift;
			my $sep		= $self->fetch_seperator();
			$pattern	=~ s/\*/.*/g;
			$pattern	=~ s/%/"[^${sep}]"/eg;
			return grep { m($pattern) } (@mailboxes, @{ $list });
		} else {
			return (@mailboxes, @{ $list });
		}
	} else {
		return undef;
	}
}

=item C<< fetch_mailbox ( $mailbox ) >>

Returns a HASHREF of status variables for the specified mailbox. Hash keys should be a subset of (MESSAGES, RECENT, UIDNEXT, UIDVALIDITY, UNSEEN).

=cut
sub fetch_mailbox {
	my $self	= shift;
	my $mailbox	= shift;
	my $rss		= $self->rss( $mailbox ) || return 0;
	my $count	= scalar(@{ $rss->{ 'items' } });
	return {
		MESSAGES	=> $count,
		RECENT		=> 0,
		UIDNEXT		=> $count+1,
		UIDVALIDITY	=> 1,
		UNSEEN		=> 0
	};
}

=item C<< list_message_uids ( $mailbox, <$from, $to> ) >>

Returns a LIST of UIDs of messages in the specified mailbox. If a range is specified with the $from and $to variables, only UIDs within the RANGE (inclusive) should be returned.

=cut
sub list_message_uids {
	my $self	= shift;
	my $mailbox	= shift;
	my $rss		= $self->rss( $mailbox ) || return 0;
	return sort { $a <=> $b } map { $self->message_time( $_ ) } $self->messages_from_rss( $rss );
}

=item C<< fetch_message ( $mailbox, $uid ) >>

Returns a LIST containing the MESSAGE and MESSAGE FIELDS specified by UID.
The MESSAGE must be either an RFC822 message object (L<Mail::Message>) or a string representation of the entire message (including header).
The MESSAGE FIELDS must be a HASHREF whose keys should be a subset of (INTERNALDATE, FLAGS).
The FLAGS value must be an ARRAYREF of flag names (e.g. 'Seen').

=cut
sub fetch_message {
	my $self	= shift;
	my $mailbox	= shift;
	my $uid		= shift;
	my $rss		= $self->rss( $mailbox ) || return 0;
	return (map { $_->[0] } grep { $_->[1] == $uid } map { [$_,$self->message_time( $_ )] } $self->messages_from_rss( $rss ))[0];
}

=item C<< fetch_body ( $mailbox, $uid ) >>

Returns either an RFC822 message body object (L<Mail::Message::Boxy>) or a string representation of the message body.

=cut
sub fetch_body {
	my $self	= shift;
	my $mailbox	= shift;
	my $uid		= shift;
	my $message	= $self->fetch_message( $mailbox, $uid );
	return $message->body;
}


=item C<< fetch_header ( $mailbox, $uid ) >>

Returns either an RFC822 message header object (L<Mail::Message::Head>) or a string representation of the message header.

=cut
sub fetch_header {
	my $self	= shift;
	my $mailbox	= shift;
	my $uid		= shift;
	my $message	= $self->fetch_message( $mailbox, $uid );
	return $message->head;
}

=item C<< fetch_fields ( $mailbox, $uid, <@fields> ) >>

If @fields are specified, returns a LIST of field values corresponding to the fields listed in @fields.
If no @fields are specified, returns a HASHREF whose keys should be a subset of (INTERNALDATE, FLAGS).
The FLAGS value must be an ARRAYREF of flag names (e.g. 'Seen').

=cut
sub fetch_fields {
	my $self	= shift;
	my $mailbox	= shift;
	my $uid		= shift;
	my $m		= $self->fetch_message( $mailbox, $uid );
	my %data	= ( FLAGS => [] );
	
	my $date	= $m->head->get('Date');
	if ($date =~ m((...), (\d\d) (...) (\d\d\d\d) (\d\d:\d\d:\d\d) (.\d\d\d\d))) {
		$data{ INTERNALDATE }	= qq("$2-$3-$4 $5 $6");
	} else {
		$data{ INTERNALDATE }	= qq("$date");
	}
	
	if (scalar(@_)) {
		return map { $data{$_} } @_;
	} else {
		return \%data;
	}
}

=item C<< add_mailbox ( $name ) >>

Adds the specified mailbox. Returns TRUE upon success, FALSE upon failure.

=item C<< remove_mailbox ( $name ) >>

Removes the specified mailbox. Returns TRUE upon success, FALSE upon failure.

=item C<< remove_message ( $mailbox, $uid ) >>

Removes the specified message. Returns TRUE upon success, FALSE upon failure.

=item C<< update_flags ( $mailbox, $uid, @flags ) >>

Updates the flags associated with the specified message with those specified.

=back

=cut



################################################################################

sub auth {
	my $self	= shift;
	my $user	= shift || '';
	my $pass	= shift || '';
	return ($self->server->data->{ 'users' }{ $user }{ 'password' } eq $pass);
}

sub count {
	my $self	= shift;
	my $mailbox	= shift;
	my $rss		= $self->rss( $mailbox ) || return 0;
	return scalar(@{ $rss->{ 'items' } });
}

sub fetch {
	my $self	= shift;
	my $mailbox	= shift;
	my $from	= shift;
	my $to		= shift;
	my $fields	= shift;
	my $by_uid	= shift;
	
	my $i			= 1;
	my $rss			= $self->rss( $mailbox ) || return [];
	my @messages	= $self->messages_from_rss( $rss );
	my %seqs		= map { reverse($i++, $_) } (reverse @messages);
	my %by_uid		= map { $self->message_time( $_ ) => $_ } @messages;
	my @uids		= sort { $a <=> $b } (keys %by_uid);
	
	if ($to eq '*') {
		$to				= ($by_uid) ? $uids[ $#uids ] : scalar(@messages);
	}
	my @results;
	my @loop	= ($by_uid) ? (grep { $_ >= $from and $_ <= $to } @uids) : ($from .. $to);
	for my $number (@loop) {
		warn "trying to get message " . $number . "/" . scalar(@messages) . "\n" if ($debug >= 2);
		my $m	= ($by_uid) ? $by_uid{ $number } : $messages[ $number - 1 ];
		next unless (ref($m));
		my $h	= $m->head;
		my $uid	= $self->message_time( $m );
		
		my $data = { _number => $seqs{ $m }, _uid => $uid };
		my %known	= map { lc($_) => 1 } $h->knownNames;
		
		warn ('_' x 30) if ($debug >= 2);
		foreach (@{ $fields }) {
			warn "fetch loop: " . Dumper($_) if ($debug >= 2);
			if (/ENVELOPE/) { 
				$data->{ ENVELOPE }	= '""';
			} elsif (/UID/) {
				$data->{ UID }	= $uid;
			} elsif (/FLAGS/) {
				$data->{ FLAGS }	= '()';
			} elsif (/INTERNALDATE/) {
				# 23-Feb-2003 01:40:20 -0800
				# Sun, 16 Mar 2003 04:20:56 -0500
				my $date	= $m->head->get('Date');
				if ($date =~ m((...), (\d\d) (...) (\d\d\d\d) (\d\d:\d\d:\d\d) (.\d\d\d\d))) {
					$data->{ INTERNALDATE }	= qq("$2-$3-$4 $5 $6");
				} else {
					$data->{ INTERNALDATE }	= qq("$date");
				}
			} elsif (/BODY/) {
				$data->{ _body }	= $m->string;
			} elsif (/RFC822\.(.*)/) {
				if ($1 eq 'SIZE') {
					$data->{ 'RFC822.SIZE' }	= $m->size;
				}
			} elsif (ref($_) and ($_->[0] eq 'BODY')) {
				my $field	= $_;
				my $headerfields	= ${ $field }[ $#{ $field } ] || [];
				
				if (grep {$_ eq 'HEADER'} @{ $field }) {
					push(@{$headerfields},keys %known) unless (@{$headerfields});
					$data->{ _body }	= join(
									'',
									map {
										ucfirst(lc($_)) . ': ' . $h->get( $_ ) . "\n"
									} grep { $known{ lc($_) } } @{ $headerfields }
								);
				} else {
					$data->{ _body }	= $m->string;
				}
			}
		}
		warn "MESSAGE DATA: " . Dumper($data) if ($debug >= 2);
		push(@results, $data);
	}
	
	return $by_uid ? \@results : [ reverse @results ];
}

sub list {
	my $self	= shift;
	my @args	= shift;
	my $list	= $self->server->user->{ 'subscriptions' };
	if (ref($list)) {
		return (@mailboxes, @{ $list });
	} else {
		return undef;
	}
}

sub status {
	my $self	= shift;
	my $mailbox	= shift;
	my $count	= $self->count( $mailbox );
	return {
		MESSAGES	=> $count,
		RECENT		=> 0,
		UIDNEXT		=> $count+1,
		UIDVALIDITY	=> 1,
		UNSEEN		=> 0
	};
}

sub subscribe {
	my $self	= shift;
	my $url		= shift;
	my %already	= map { $_ => 1 } @{ $self->server->user->{ 'subscriptions' } };
	
	if (my $rss = $self->rss( $url )) {
		push( @{ $self->server->user->{ 'subscriptions' } }, $url ) unless ($already{ $url });
		return 1;
	} else {
		return 0;
	}
}

sub unsubscribe {
	my $self	= shift;
	my $url		= shift;
	my $subs	= $self->server->user->{ 'subscriptions' };
	@{ $subs }	= grep { $_ ne $url } @{ $subs };
	return 1;
}

################################################################################

sub get_url {
	my $self	= shift;
	my $url		= shift;
	
# CACHING is taken care of in rss() - look there instead
#	if (my $cache = $self->_feed_cache->{ $url }[1]) {
#		return $cache;
#	} else {
		my $response	= $self->ua->get( $url );
		if ($response->is_success) {
			$self->{ _feed_cache }{ $url }	= [ time, $response->content ];
			return $response->content;
		} else {
			return undef;
		}
#	}
}

sub rss {
	my $self	= shift;
	my $url		= shift;
	
	$url		= $DEFAULT_URL if ($url eq 'INBOX');
	
	no warnings 'uninitialized';
	if (ref(my $rss = $self->_feed_cache->{ $url }[2]) and (time - $self->_feed_cache->{ $url }[0] < ($CACHE_MINUTES * 60))) {
		# IF we have the RSS file cached, AND we've updated within the
		# last $CACHE_MINUTES minutes THEN use the cached RSS object 
		warn "got cached rss feed";
		return $rss;
	} else {
		my $rss		= new XML::RSS;
		my @urls	= ( $url );
		while (my $rssurl = shift(@urls)) {
			warn "fetching << $rssurl >> looking for RSS" if ($debug);
			my $content	= $self->get_url( $rssurl );
			if (defined($content)) {
				eval { $rss->parse( $content ) };
				
				unless ($@) {
					warn "-> got parsable rss feed";
					$self->_feed_cache->{ $url }[2]	= $rss;
					return $rss;
				} else {
					my $html	= HTML::RSSAutodiscovery->new();
					my $alts	= $html->parse( $rssurl );
					warn "looking for rss feeds at << $rssurl >>";
					if (ref($alts) and scalar(@{ $alts })) {
						foreach (@{ $alts }) {
							next unless ($_->{ 'type' } eq 'application/rss+xml');
							my $href	= $_->{ 'href' };
							warn "-> found potential rss feed << $href >>";
							if ($href =~ m(^http(s)?://)) {
								warn "found RSS at << $href >>";
								push(@urls, $href);
							} else {
								$href	= substr( $rssurl, 0, rindex($rssurl,'/') + 1 ) . $href;
								push(@urls, $href);
							}
						}
					} else {
						warn "-> no rss feeds found";
					}
				}
			} else {
				warn "-> No content found";
			}
		}
		
		# something went wrong...
		warn "No RSS feed at $url\n" if ($debug);
		return undef;
	}
}

sub messages_from_rss {
	my $self	= shift;
	my $rss		= shift;
	
	my @messages;
	my $title	= $rss->{ 'channel' }{ 'title' };
	my $format	= HTML::FormatText->new(leftmargin => 0, rightmargin => 70);
	
	foreach my $item (@{ $rss->{ 'items' } }) {
		my $content	= (exists($item->{'http://purl.org/rss/1.0/modules/content/'}{'encoded'}))
				? $item->{'http://purl.org/rss/1.0/modules/content/'}{'encoded'}
				: $item->{'description'};
		my $tree	= HTML::TreeBuilder->new->parse($content); 
		
		my $html	= new Mail::Message::Body ( mime_type => 'text/html', data => $content );
		my $plain	= new Mail::Message::Body ( mime_type => 'text/plain', data => $format->format($tree) );
		my $mpbody	= new Mail::Message::Body::Multipart ( mime_type => 'multipart/alternative', preamble => Mail::Message::Body->new( data => "This is a multi-part message in MIME format.\n\n") , parts => [ $plain, $html ] );
		my $m		= Mail::Message->buildFromBody ( $mpbody,
					# 'RSS-IMAP Bridge <greg@evilfunhouse.com>',
					To 			=> join(' ', $title, '<rss+bridge@evilfunhouse.com>'),
					From		=> join(' ', $title, '<rss+bridge@evilfunhouse.com>'),
					Subject		=> $item->{ 'title' },
				);
		if (my $date = $item->{'dc'}{'date'}) {
			# 2003-02-24T02:39:11-05:00
			my ($year,$mon,$day,$hour,$min,$sec,$o_dir,$o_hour,$o_min)
				= $date =~ m{^(\d{4})-(\d{1,2})-(\d{1,2}).(\d{1,2}):(\d{2}):(\d{1,2})(.)(\d{1,2}):(\d{1,2})$};
			my $time	= timegm( $sec, $min, $hour, $day, $mon-1, $year-1900 );
			$time		+= (($o_hour * 60 * 60) + ($o_min * 60)) * (($o_dir eq '-') ? 1 : -1);
			my $dstr	= strftime( "%a, %d %b %Y %H:%M:%S", (localtime($time))[0..6] )
						. " -${o_hour}${o_min}";
			$m->head->delete( 'Date' );
			$m->head->add( Date => $dstr );
		}
		push(@messages, $m);
	}
	return @messages;
}

{
	my %months	= (
		Jan	=> 0,
		Feb	=> 1,
		Mar	=> 2,
		Apr	=> 3,
		May	=> 4,
		Jun	=> 5,
		Jul	=> 6,
		Aug	=> 7,
		Sep	=> 8,
		Oct	=> 9,
		Nov	=> 10,
		Dec	=> 11
	);
	sub message_time {
		my $self	= shift;
		my $m		= shift;
		my $date	= $m->head->get( 'Date' );
		my ($day,$mon,$year,$hour,$min,$sec)	= $date =~ /^..., (..) (...) (....) (..):(..):(..)/;
		my $time	= timegm( $sec, $min, $hour, $day, $months{$mon}, $year-1900);
		return $time;
	}
}

################################################################################

sub AUTOLOAD {
	our $AUTOLOAD;
	my $self	= shift;
	my $name	= substr($AUTOLOAD,rindex($AUTOLOAD,':')+1);
	return if ($name eq 'DESTROY');
	if (scalar(@_)) {
		$self->{ $name }	= shift;
	} else {
		return $self->{ $name };
	}
}

1;

__END__

=head1 REVISION HISTORY

 $Log: RSSBridge.pm,v $
 Revision 1.6  2003/03/25 08:08:24  greg
 - add support for proper INTERNALDATE format
 - add initial code for cleaner backend API (not used yet)
 - tweaks to fetch() to make Outlook happy (not sure if it still is)

 Revision 1.5  2003/03/17 09:16:40  greg
 - added initial support for RSS autodiscovery
 - added MIME alternative encodings for HTML and plain text
   currently needs patch to Mail::Message::Body::Multipart::init() :
   http://marc.theaimsgroup.com/?l=perl-mailbox&m=104788682531815&w=2

 Revision 1.4  2003/03/17 07:18:32  greg
 - added first cut at a tokenizer for FETCH commands (no more regex parsing)
 - moved {adduser,imap}.pl to bin/ and updated MANIFEST
 - changed name in Makefile.PL to 'Net::IMAP::Server'
 - added debugging info

 Revision 1.3  2003/03/03 14:48:12  greg
 - indented Log tags for (future) perldoc handling

 Revision 1.2  2003/03/03 14:10:14  greg
 - added PREREQ_PM to makefile
 - command tag is now an instance variable
 - command tag is no longer passed to cmd_* and send_* functions
 - rid code of unused Switch references
 - added CVS Revision, Date, and Log tags

=cut
